import os
import py_compile


class FileAnalyzer:
    def __init__(self, filename):
        self.filename = filename
        self.message = None
    def file_exists(self):
        return os.path.isfile(self.filename)

    def file_not_empty(self):
        return self.file_exists() and os.path.getsize(self.filename) > 0

    def file_is_python(self):
        if not self.filename.endswith(".py"):
            return False
        try:
            py_compile.compile(self.filename, doraise=True)
            return True
        except (SyntaxError, py_compile.PyCompileError) as e:
            self.message = str(e)
            return False



import sys, types

class OutOfInputsError(RuntimeError):
    pass


class _ListenerStream:
    def __init__(self, data_listener):
        self._data_listener = data_listener


class _InputListenerStream(_ListenerStream):

    def __init__(self, data_listener, inputs_iterator):
        _ListenerStream.__init__(self, data_listener)
        self._inputs_iterator = inputs_iterator

    def fetch_remaining_inputs(self):
        result = []
        while True:
            try:
                result.append(next(self._inputs_iterator))
            except StopIteration:
                break

        return result

    def read(self, limit=-1):
        raise NotImplementedError("The method 'read' on stdin is not supported in testable programs")

    def readline(self, limit=-1):
        if limit > 0:
            raise NotImplementedError("Positive limit/size for method 'readline' on stdin "
                                      "is not supported in testable programs")

        try:
            data = str(next(self._inputs_iterator)) + "\n"
            self._data_listener(data)
            return data
        except StopIteration:
            raise OutOfInputsError("The program asked for more inputs than were available")

    def readlines(self, limit=-1):
        raise NotImplementedError("The method 'readlines' on stdin is not supported in testable programs")


class _OutputListenerStream(_ListenerStream):
    def __init__(self, data_listener):
        _ListenerStream.__init__(self, data_listener)

    def write(self, data):
        self._data_listener(data)

    def writelines(self, lines):
        self._data_listener(lines)


class IOCapturing:
    def __init__(self, inputs=[]):
        self._inputs = inputs

    def __enter__(self):
        self._inputs_iterator = self._create_inputs_iterator(self._inputs)
        self._stream_events = []
        self._original_streams = {}
        self._remaining_inputs = []

        # Remember original streams
        for stream_name in {"stdin", "stdout", "stderr"}:
            self._original_streams[stream_name] = getattr(sys, stream_name)

        # Install fake streams
        sys.stdin = _InputListenerStream(self._record_stdin_data, self._inputs_iterator)
        sys.stdout = _OutputListenerStream(self._record_stdout_data)
        sys.stderr = _OutputListenerStream(self._record_stderr_data)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._remaining_inputs = sys.stdin.fetch_remaining_inputs()

        # restore original streams
        for stream_name in {"stdin", "stdout", "stderr"}:
            setattr(sys, stream_name, self._original_streams[stream_name])

    def _create_inputs_iterator(self, inputs):
        if isinstance(inputs, str):
            return iter(inputs.splitlines())

        try:
            return iter(inputs)
        except TypeError:
            try:
                result = inputs(self)
                assert isinstance(result, types.GeneratorType)
                return result
            except BaseException as e:
                e.message = "Could not create iterator for inputs.\n" + e.message
                raise

    def _record_stdin_data(self, data):
        self._stream_events.append(("stdin", data))

    def _record_stdout_data(self, data):
        self._stream_events.append(("stdout", data))

    def _record_stderr_data(self, data):
        self._stream_events.append(("stderr", data))

    def get_remaining_inputs(self):
        return self._remaining_inputs

    def _get_stream_data(self, stream_names, only_since_last_input):
        result = ""
        i = len(self._stream_events) - 1
        while i >= 0:
            event_stream_name, event_data = self._stream_events[i]

            if only_since_last_input and event_stream_name == "stdin":
                break

            if event_stream_name in stream_names:
                result = event_data + result

            i -= 1

        return result

    def get_io(self):
        return self._get_stream_data({"stdout", "stderr", "stdin"}, False)

    def get_stdout(self):
        return self._get_stream_data({"stdout"}, False)

    def get_last_stdout(self):
        return self._get_stream_data({"stdout"}, True)

    def get_stderr(self):
        return self._get_stream_data({"stderr"}, False)

    def debug(self, *args, sep=' ', end='\n', stream_name="stdout", flush=False):
        """Meant for printing debug information from input generator."""
        print(*args, sep=sep, end=end,
              file=self._original_streams[stream_name], flush=flush)




from typing import List

from tiivad.execution_analyzer import *
from tiivad.file_analyzer import *
from tiivad.results import Results
from tiivad.syntax_tree_analyzer import *

STATIC_TESTS_ONE = {
    'program_contains_loop_test',
    'program_contains_try_except_test',
    'program_calls_print_test',
    'function_contains_loop_test',
    'function_contains_try_except_test',
    'function_calls_print_test',
    'function_contains_return_test',
    'function_is_pure_test',
    'function_is_recursive_test'
}
STATIC_TESTS_MANY = {
    'program_imports_module_test',
    'program_defines_function_test',
    'program_calls_function_test',
    'program_contains_keyword_test',
    'program_defines_class_test',
    'program_calls_class_test',
    'program_calls_class_function_test',
    'class_imports_module_test',
    'class_defines_function_test',
    'class_calls_class_test',
    'function_imports_module_test',
    'function_defines_function_test',
    'function_calls_function_test',
    'function_contains_keyword_test'
}
EXECUTION_TESTS = {
    'program_execution_test',
    'class_instance_test',
    'function_execution_test'
}


class TestResult:
    PASS = "PASS"
    FAIL = "FAIL"


def format_message(msg: str, d: dict):
    if d is None or d == {}:
        return msg
    for k, v in d.items():
        if "{" + str(k) + "}" in msg:
            msg = msg.replace("{" + str(k) + "}", str(v))
    return msg


def check_result(title, status, feedback, format_dict=None):
    return {'title': format_message(title, format_dict),
            'status': status,
            'feedback': format_message(feedback, format_dict)}


def validate_files(filenames: List[str]) -> bool:
    for filename in filenames:
        fa = FileAnalyzer(filename)
        if not fa.file_exists():
            Results.pre_evaluate_error = f"Faili {filename} ei ole."
            return False
        elif not fa.file_not_empty():
            Results.pre_evaluate_error = f"Fail {filename} on tühi."
            return False
        elif not fa.file_is_python():
            if fa.message is None:
                Results.pre_evaluate_error = f"Fail {filename} ei ole korrektne Pythoni fail."
            else:
                Results.pre_evaluate_error = fa.message
            return False
    return True


def execute_test(**kwargs):
    if Results.pre_evaluate_error:
        return

    test_type = kwargs["type"]
    component, check_type = test_type.split("_", 1)
    check_type = check_type[:-len("_test")]

    checks = []
    test_exception_message = None
    actual_output = None
    converted_submission = None
    test_status = TestResult.PASS

    try:
        test_status, actual_output, converted_submission = run_test(
            check_type,
            checks,
            component,
            kwargs,
            test_status,
            test_type
        )
    except Exception as e:
        test_status = TestResult.FAIL
        test_exception_message = str(e)

    Results.total_points += kwargs["points_weight"]
    if test_status == TestResult.PASS:
        Results.passed_points += kwargs["points_weight"]
    Results({
        "title": kwargs.get("name", ""),
        "user_inputs": kwargs.get("standard_input_data", []),
        "created_files": [{"name": x[0], "content": x[1]} for x in kwargs.get("input_files", [])],
        "converted_submission": converted_submission,
        "actual_output": actual_output,
        "exception_message": test_exception_message,
        "status": test_status,
        "checks": checks if test_exception_message is None else []
    })


def run_test(check_type, checks, component, kwargs, test_status, test_type):
    actual_output = None
    converted_submission = None

    if test_type in STATIC_TESTS_ONE | STATIC_TESTS_MANY:
        if component == "program":
            ta = ProgramSyntaxTreeAnalyzer(kwargs["file_name"])
        elif component == "class":
            ta = ClassSyntaxTreeAnalyzer(kwargs["file_name"], kwargs["class_name"])
        elif component == "function":
            ta = FunctionSyntaxTreeAnalyzer(kwargs["file_name"], kwargs["function_name"])
        else:
            ta = None

        for check in kwargs.get("generic_checks", []) + kwargs.get("contains_checks", []):
            if ta is not None and ta.tree is not None and \
                    (test_type in STATIC_TESTS_ONE and getattr(ta, check_type)() == check['expected_value'] or
                     test_type in STATIC_TESTS_MANY and ta.analyze_with_quantifier(check_type, check['check_type'],
                                                                                   set(check['expected_value']),
                                                                                   check['nothing_else'])):
                checks.append(check_result(check['before_message'], TestResult.PASS, check['passed_message'],
                                           ta.__dict__))
            else:
                checks.append(check_result(check['before_message'], TestResult.FAIL, check['failed_message'],
                                           ta.__dict__))
                test_status = TestResult.FAIL
                break

    elif test_type in EXECUTION_TESTS:
        if component == "program":
            ea = ProgramExecutionAnalyzer(kwargs["file_name"], kwargs.get("standard_input_data", []),
                                          kwargs.get("input_files", []))
        elif component == "class":
            ea = ClassExecutionAnalyzer(kwargs["file_name"], kwargs.get("create_object", ""),
                                        kwargs.get("standard_input_data", []), kwargs.get("input_files", []))
        elif component == "function":
            ea = FunctionExecutionAnalyzer(kwargs["file_name"], kwargs["function_name"], kwargs["function_type"],
                                           kwargs.get("create_object", "pass"), kwargs["arguments"],
                                           kwargs.get("standard_input_data", []), kwargs.get("input_files", []))
        else:
            ea = None

        actual_output = ea.all_output
        converted_submission = ea.converted_script

        # Erind
        if ea.raised_exception() and isinstance(ea.exception, OutOfInputsError):
            test_status = TestResult.FAIL
            checks.append(
                check_result(test_type, test_status, "Programm küsis rohkem sisendeid, kui testil oli anda"))
        elif 'exception_check' in kwargs and kwargs['exception_check'] is not None:
            check = kwargs['exception_check']
            if ea.raised_exception() == check.get('expected_value', False):
                message = check['passed_message']
            else:
                test_status = TestResult.FAIL
                message = check['failed_message']
            checks.append(check_result(check['before_message'], test_status, message, ea.__dict__))
        elif ea.raised_exception():
            test_status = TestResult.FAIL
            checks.append(
                check_result(test_type, test_status, f"{type(ea.exception).__name__}: {str(ea.exception)}"))

        check_lists = [
            # Standardväljund ja väljundfail
            kwargs.get("standard_output_checks", []) + kwargs.get("output_file_checks", []),
            # Funktsiooni tagastusväärtus ja muudetud väärtus
            kwargs.get("return_value_checks", []) + kwargs.get("param_value_checks", []),
            # Objekti olek
            kwargs.get("class_instance_checks", [])
        ]

        for i, check_list in enumerate(check_lists):
            if test_status == TestResult.FAIL:
                break

            for check in check_list:
                if i == 0:
                    result = ea.analyze_output_with_quantifier(check.get('file_name', None), check['data_category'],
                                                               check['check_type'], check['expected_value'],
                                                               check['nothing_else'], check['elements_ordered'])
                elif i == 1:
                    result = ea.value_correct(check.get('param_number', None), check['expected_value'])
                elif i == 2:
                    result = ea.fields_correct(check['fields_final'], check['check_name'], check['check_value'],
                                               check['nothing_else'])
                else:
                    result = False

                if result:
                    checks.append(check_result(check['before_message'], TestResult.PASS, check['passed_message'],
                                               ea.__dict__))
                else:
                    test_status = TestResult.FAIL
                    checks.append(check_result(check['before_message'], TestResult.FAIL, check['failed_message'],
                                               ea.__dict__))
                    break

    return test_status, actual_output, converted_submission


if __name__ == "__main__":
    execute_test(file_name='''../test/samples/prog5-raamat.py''',
                 type='''class_instance_test''',
                 create_object='''return Raamat("Kogumik", 12.5)''',
                 standard_input_data=[],
                 input_files=[],
                 class_instance_checks=[
                     {'fields_final': [('''nimi''', '''Kogumik'''), ('''hind''', 12.5)],
                      'check_name': True,
                      'check_value': True,
                      'nothing_else': True,
                      'before_message': '''Kas objekti väljad on õiged?''',
                      'passed_message': '''Objekti väljad on õiged.''',
                      'failed_message': '''Objekti väljad ei ole õiged.'''
                      },
                     {'fields_final': [('''nimi''', '''Abc'''), ('''hind''', 1.23)],
                      'check_name': True,
                      'check_value': False,
                      'nothing_else': True,
                      'before_message': '''Kas objekti väljad on õiged?''',
                      'passed_message': '''Objekti väljad on õiged.''',
                      'failed_message': '''Objekti väljad ei ole õiged.'''
                      },
                     {'fields_final': [('''nimi''', '''Abc'''), ('''hind''', 1.23), ('''autor''', '''X''')],
                      'check_name': True,
                      'check_value': False,
                      'nothing_else': True,
                      'before_message': '''Kas objekti väljad on õiged?''',
                      'passed_message': '''Objekti väljad on õiged.''',
                      'failed_message': '''Objekti väljad ei ole õiged.'''
                      }
                 ],
                 points_weight=1
                 )
    print(Results(None))



import ast
import re


class ValidationType:
    ALL_OF_THESE = "ALL_OF_THESE"
    ANY_OF_THESE = "ANY_OF_THESE"
    ANY = "ANY"
    NONE_OF_THESE = "NONE_OF_THESE"
    MISSING_AT_LEAST_ONE_OF_THESE = "MISSING_AT_LEAST_ONE_OF_THESE"
    NONE = "NONE"


class ProgramSyntaxTreeAnalyzer:
    def __init__(self, program_name, class_name=None, function_name=None):

        self.imports_module_names, self.defines_function_names = set(), set()
        self.defines_class_names, self.defines_subclass_names = set(), set()
        self.calls_function_names, self.calls_class_function_names = set(), set()
        self.contains_keyword_names, self.defined_vars = set(), set()
        self.contains_loop_tv = self.contains_try_except_tv = self.contains_return_tv = False
        self.is_class_tv = self.is_function_tv = self.is_pure_tv = False

        try:
            with open(program_name, encoding="utf-8") as f:
                self.tree = ast.parse(f.read())
        except:
            self.tree = None

        for node_type, name in [(ast.ClassDef, class_name), (ast.FunctionDef, function_name)]:
            if self.tree is not None and name is not None:
                self.tree = next((x for x in ast.walk(self.tree)
                                  if isinstance(x, node_type) and x.name == name), None)

        if self.tree is None:
            return

        self.is_class_tv = isinstance(self.tree, ast.ClassDef)
        self.is_function_tv = self.is_pure_tv = isinstance(self.tree, ast.FunctionDef)
        self.contains_keyword_names = set(re.findall(r'\w+', ast.unparse(self.tree)))

        for x in ast.walk(self.tree):
            node_type = type(x).__name__
            if x == self.tree:
                continue
            if node_type == 'Import':
                for y in x.names:
                    self.imports_module_names |= set(y.name.split("."))
            elif node_type == 'ImportFrom':
                self.imports_module_names |= set(x.module.split("."))
            elif node_type == 'ClassDef':
                self.defines_class_names.add(x.name)
                for y in x.bases:
                    self.defines_subclass_names.add((x.name, y.id))
            elif node_type == 'FunctionDef':
                self.defines_function_names.add(x.name)
            elif node_type in ['For', 'While', 'comprehension']:
                self.contains_loop_tv = True
            elif node_type in ['Try', 'ExceptHandler']:
                self.contains_try_except_tv = True
            elif node_type == 'Call':
                if isinstance(x.func, ast.Name):
                    self.calls_function_names.add(x.func.id)
                    self.defined_vars.add(x.func.id)
                elif isinstance(x.func, ast.Attribute):
                    self.calls_function_names.add(x.func.attr)
                    self.calls_class_function_names.add(x.func.attr)
                    self.defined_vars.add(x.func.attr)
            elif node_type == 'arg':
                self.defined_vars.add(x.arg)
            elif node_type == 'Name':
                if isinstance(x.ctx, ast.Store):
                    self.defined_vars.add(x.id)
                elif isinstance(x.ctx, ast.Load) and x.id not in self.defined_vars and \
                        x.id not in self.imports_module_names:
                    self.is_pure_tv = False
            elif node_type == 'Return':
                self.contains_return_tv = True

    def is_class(self) -> bool:
        return self.is_class_tv

    def is_function(self) -> bool:
        return self.is_function_tv

    def imports_module(self, name: str = None) -> bool:
        return len(self.imports_module_names) > 0 if name is None \
            else name in self.imports_module_names

    def defines_class(self, name: str = None) -> bool:
        return len(self.defines_class_names) > 0 if name is None \
            else name in self.defines_class_names

    def defines_function(self, name: str = None) -> bool:
        return len(self.defines_function_names) > 0 if name is None \
            else name in self.defines_function_names

    def contains_loop(self) -> bool:
        return self.contains_loop_tv

    def contains_try_except(self) -> bool:
        return self.contains_try_except_tv

    def contains_keyword(self, name: str = None) -> bool:
        return len(self.contains_keyword_names) > 0 if name is None \
            else name in self.contains_keyword_names

    def calls_function(self, name: str = None) -> bool:
        return len(self.calls_function_names) > 0 if name is None \
            else name in self.calls_function_names

    def calls_print(self) -> bool:
        return "print" in self.calls_function_names

    def creates_instance(self, name: str = None) -> bool:
        return len(self.calls_function_names & self.defines_class_names) > 0 if name is None \
            else name in self.calls_function_names & self.defines_class_names

    def calls_class_function(self, name: str = None) -> bool:
        return len(self.calls_class_function_names) > 0 if name is None \
            else name in self.calls_class_function_names

    def analyze_with_quantifier(self, target: str, quantifier: str, names: set = set(),
                                nothing_else: bool = False) -> bool:
        match target:
            case 'imports_module':
                targetset = self.imports_module_names
            case 'defines_function':
                targetset = self.defines_function_names
            case 'calls_function':
                targetset = self.calls_function_names
            case 'contains_keyword':
                targetset = self.contains_keyword_names
            case 'defines_class':
                targetset = self.defines_class_names
            case 'creates_instance':
                targetset = self.defines_class_names & self.calls_function_names
            case 'calls_class_function':
                targetset = self.calls_class_function_names
            case _:
                return False
        match quantifier:
            case ValidationType.ALL_OF_THESE:
                return names <= targetset and (not nothing_else or targetset <= names)
            case ValidationType.ANY_OF_THESE:
                return len(names & targetset) > 0 and (not nothing_else or targetset <= names)
            case ValidationType.ANY:
                return len(targetset) > 0
            case ValidationType.MISSING_AT_LEAST_ONE_OF_THESE:
                return not (names <= targetset)
            case ValidationType.NONE_OF_THESE:
                return len(names & targetset) == 0
            case ValidationType.NONE:
                return len(targetset) == 0
            case _:
                return False


class ClassSyntaxTreeAnalyzer(ProgramSyntaxTreeAnalyzer):
    def __init__(self, program_name, class_name):
        super().__init__(program_name, class_name)


class FunctionSyntaxTreeAnalyzer(ProgramSyntaxTreeAnalyzer):
    def __init__(self, program_name, function_name):
        super().__init__(program_name, None, function_name)

    def is_pure(self) -> bool:
        return self.is_pure_tv

    def contains_return(self) -> bool:
        return self.contains_return_tv

    def is_recursive(self) -> bool:
        return self.is_function_tv and self.calls_function(self.tree.name)

    def prints_instead_of_returning(self) -> bool:
        return self.is_function_tv and self.calls_print() and not self.contains_return()


if __name__ == "__main__":
    ca = FunctionSyntaxTreeAnalyzer("../test/samples/prog3-calc.py", "arv")
    print(ca.defined_vars, ca.imports_module_names)
    print(ca.is_pure())



from tiivad.version import __version__
import json


class Results(object):
    __instance = None
    pre_evaluate_error = None
    total_points = 0
    passed_points = 0
    tests = []

    def __new__(cls, val):
        if Results.__instance is None:
            Results.__instance = object.__new__(cls)
        if val is not None:
            Results.__instance.tests.append(val)
        return Results.__instance

    def __str__(cls):
        d = {
            'producer': f'tiivad {__version__}',
            'result_type': 'OK_V3',
            'pre_evaluate_error': cls.pre_evaluate_error,
            'points': round(100 * cls.passed_points / cls.total_points) if cls.total_points > 0 else 0,
            'tests': cls.tests
        }
        return json.dumps(d, ensure_ascii=False)

    def clear(cls):
        cls.__instance = None
        cls.total_points = 0
        cls.passed_points = 0
        cls.tests = []
        cls.pre_evaluate_error = None


if __name__ == "__main__":
    Results.passed_points = 2
    Results.total_points = 3
    print(Results(1))



from setuptools import setup, find_namespace_packages
exec(open('tiivad/version.py').read())

setup(
    name='tiivad',
    version=__version__,
    license='MIT',
    author="Eerik Muuli",
    author_email='eerik.muuli@gmail.com',
    packages=find_namespace_packages(),
    url='https://github.com/emuuli/tsl-tiivad',
    install_requires=[
    ],

)



import ast
import re
import runpy

from tiivad.capture_io import IOCapturing, OutOfInputsError
from tiivad.syntax_tree_analyzer import ValidationType


class ElementsType:
    CONTAINS_STRINGS = "CONTAINS_STRINGS"
    CONTAINS_LINES = "CONTAINS_LINES"
    CONTAINS_NUMBERS = "CONTAINS_NUMBERS"
    EQUALS = "EQUALS"


def create_file(file_name, file_content, text_file_encoding=None):
    if text_file_encoding is None:
        text_file_encoding = "UTF-8"

    if isinstance(file_content, str):
        file_content = file_content.encode(text_file_encoding)

    with open(file_name, mode="wb") as fp:
        fp.write(file_content)


def convert_script(file_name, converted_file_name, create_object_body=None):
    with open(file_name, encoding="utf8") as f:
        file_content = f.read()
    lines = file_content.splitlines()
    try:
        tree = ast.parse(file_content)
    except:
        pass
    else:
        if 'body' in tree._fields:
            for node in tree.body:
                comment = True
                if isinstance(node, (ast.FunctionDef, ast.ClassDef, ast.Import, ast.ImportFrom, ast.Assign)):
                    comment = False
                    if isinstance(node, ast.Assign):
                        for x in ast.walk(node.value):
                            if isinstance(x, (ast.Call, ast.Name)):
                                comment = True
                if comment:
                    for i in range(node.lineno - 1, node.end_lineno):
                        lines[i] = '#' + lines[i]
    body = '\n'.join(lines)
    if create_object_body:
        body += '\ndef create_object_fun_auto_assess():'
        body += ''.join('\n    ' + line for line in create_object_body.splitlines())
    with open(converted_file_name, 'w', encoding="utf8") as f:
        f.write(body)
    return body


def extract_numbers(s):
    numbers = []
    for el in s.split():
        el = el.rstrip(".,:;?!)]}'\"%€$£")
        if re.match(r"-?(?!0\d+)\d+$", el):
            numbers.append(int(el))
        elif re.match(r"-?(?!0\d+)\d+\.\d+$", el):
            numbers.append(float(el))
    return numbers


def extract_strings(s, word_list):
    e = r"\b(" + "|".join(word_list) + r")\b" if word_list else None
    return [] if e is None else re.findall(e, s)


def extract_lines(s):
    lines = s.splitlines()
    return [line for line in lines if line.strip()]


class ProgramExecutionAnalyzer:
    def __init__(self, file_name: str, user_inputs: list, input_files: set):
        for name, content in input_files:
            create_file(name, content)
        with IOCapturing(user_inputs) as iocap:
            self.exception = None
            try:
                self.globals_dict = runpy.run_path(file_name, run_name="__main__")
            except Exception as e:
                self.exception = e
        if isinstance(self.exception, OutOfInputsError):
            self.actual_input_count = float('Inf')
        else:
            self.actual_input_count = len(user_inputs) - len(iocap.get_remaining_inputs())
        self.all_io = iocap.get_io()
        self.all_output = iocap.get_stdout()
        self.last_output = iocap.get_last_stdout()
        self.converted_script = None

    def raised_exception(self) -> bool:
        return self.exception is not None

    def analyze_output_with_quantifier(self, file_name, data_category: str, quantifier, values: list,
                                       nothing_else: bool = False, ordered: bool = False) -> bool:
        if file_name:
            with open(file_name) as f:
                content = f.read()
        else:
            content = self.last_output

        match data_category:
            case ElementsType.CONTAINS_NUMBERS:
                all_values = extract_numbers(content)
            case ElementsType.CONTAINS_STRINGS:
                all_values = extract_strings(content, values)
            case ElementsType.CONTAINS_LINES:
                all_values = extract_lines(content)
            case ElementsType.EQUALS:
                all_values = [content.rstrip()]
            case _:
                all_values = []
        match quantifier:
            case ValidationType.ALL_OF_THESE:
                if not ordered:
                    values.sort()
                    all_values.sort()
                i = 0
                for value in all_values:
                    if values[i] == value:
                        i += 1
                        if i == len(values):
                            break
                return i == len(values) and (not nothing_else or len(values) == len(all_values))
            case ValidationType.ANY_OF_THESE:
                return len(set(values) & set(all_values)) > 0 and (not nothing_else or set(all_values) <= set(values))
            case ValidationType.ANY:
                return len(all_values) > 0
            case ValidationType.MISSING_AT_LEAST_ONE_OF_THESE:
                return not (set(values) <= set(all_values))
            case ValidationType.NONE_OF_THESE:
                return len(set(values) & set(all_values)) == 0
            case ValidationType.NONE:
                return len(all_values) == 0
            case _:
                return False

    def analyze_exception(self, target: str) -> bool:
        pass


class ClassExecutionAnalyzer(ProgramExecutionAnalyzer):
    def __init__(self, file_name: str, create_object_body: str, user_inputs: list = [], input_files: set = set()):
        converted_file_name = "converted_file.py"
        self.converted_script = convert_script(file_name, converted_file_name, create_object_body)
        super().__init__(converted_file_name, [], input_files)
        if self.exception is not None:
            return

        for name, content in input_files:
            create_file(name, content)
        with IOCapturing(user_inputs) as iocap:
            try:
                self.obj = self.globals_dict.get("create_object_fun_auto_assess", None)()
            except Exception as e:
                self.exception = e

        if isinstance(self.exception, OutOfInputsError):
            self.actual_input_count = float('Inf')
        else:
            self.actual_input_count = len(user_inputs) - len(iocap.get_remaining_inputs())
        self.all_io = iocap.get_io()
        self.all_output = iocap.get_stdout()
        self.last_output = iocap.get_last_stdout()

    def fields_exist(self, quantifier: str, names: set, nothing_else: bool = False) -> bool:
        targetset = set(self.obj.__dict__.keys())
        match quantifier:
            case ValidationType.ALL_OF_THESE:
                return names <= targetset and (not nothing_else or targetset <= names)
            case ValidationType.ANY_OF_THESE:
                return len(names & targetset) > 0 and (not nothing_else or targetset <= names)
            case ValidationType.ANY:
                return len(targetset) > 0
            case ValidationType.MISSING_AT_LEAST_ONE_OF_THESE:
                return not (names <= targetset)
            case ValidationType.NONE_OF_THESE:
                return len(names & targetset) == 0
            case ValidationType.NONE:
                return len(targetset) == 0
            case _:
                return False

    def fields_correct(self, fields: list, check_name: bool, check_value: bool, nothing_else: bool = False) -> bool:
        for f in fields:
            if not any((not check_name or f[0] == of) and (not check_value or f[1] == ov)
                       for of, ov in self.obj.__dict__.items()):
                return False
        if nothing_else:
            for of, ov in self.obj.__dict__.items():
                if not any((not check_name or f[0] == of) and (not check_value or f[1] == ov)
                           for f in fields):
                    return False
        return True

    def obj_to_str(self) -> str:
        if hasattr(self.obj, '__dict__'):
            s = [f"{k} = {repr(v)}" for k, v in self.obj.__dict__.items()]
            return self.obj.__class__.__name__ + "(" + ", ".join(s) + ")"
        else:
            return str(self.obj)


class FunctionExecutionAnalyzer(ProgramExecutionAnalyzer):
    def __init__(self, file_name: str, function_name: str, function_type: str, create_object_body: str,
                 arguments: list, user_inputs: list = [], input_files: set = set()):
        self.arguments = arguments
        self.result = None
        converted_file_name = "converted_file.py"
        self.converted_script = convert_script(file_name, converted_file_name, create_object_body)
        super().__init__(converted_file_name, [], input_files)
        if self.exception is not None:
            return

        if function_type == "FUNCTION":
            f_obj = self.globals_dict.get(function_name, None)
        else:  # "METHOD"
            try:
                obj = self.globals_dict.get("create_object_fun_auto_assess", None)()
                f_obj = getattr(obj, function_name)
            except Exception as e:
                self.exception = e
                return

        for name, content in input_files:
            create_file(name, content)
        with IOCapturing(user_inputs) as iocap:
            try:
                self.result = f_obj(*arguments)
            except Exception as e:
                self.exception = e
        if isinstance(self.exception, OutOfInputsError):
            self.actual_input_count = float('Inf')
        else:
            self.actual_input_count = len(user_inputs) - len(iocap.get_remaining_inputs())
        self.all_io = iocap.get_io()
        self.all_output = iocap.get_stdout()
        self.last_output = iocap.get_last_stdout()

    def value_correct(self, param_number, value) -> bool:
        if param_number is None:
            return self.result == value
        else:
            return param_number < len(self.arguments) and self.arguments[param_number] == value


if __name__ == "__main__":
    file_name = f"../test/samples/func_exec1.py"
    function_name = "muuda_järjendit"
    function_type = "FUNCTION"
    create_object_body = None
    arguments = [2, [3, 4, 5]]
    user_inputs = []
    input_files = set()
    ea = FunctionExecutionAnalyzer(file_name, function_name, function_type, create_object_body, arguments,
                                   user_inputs, input_files)
    print(ea.all_io)
    print(ea.result)
    print(ea.exception)
    print(arguments)
    print(ea.value_correct(1, [3, 4, 1]))
